﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;

using System.Drawing.Drawing2D;

namespace howto_flow_blocks
{
    public partial class Form1 : Form
    {
        public Form1()
        {
            InitializeComponent();
        }

        // Obstacles.
        private List<RectangleF> Obstacles;

        // Blocks to flow around obstacles.
        private List<Block> Blocks;

        // Define some blocks.
        private void Form1_Load(object sender, EventArgs e)
        {
            Obstacles = new List<RectangleF>();
            Obstacles.Add(new RectangleF(0, 0, 64, 64));
            Obstacles.Add(new RectangleF(200, 50, 50, 100));
            Obstacles.Add(new RectangleF(50, 150, 100, 70));
            Obstacles.Add(new RectangleF(300, 200, 100, 64));

            // Random rand = new Random(4);
            Random rand = new Random(4);
            Blocks = new List<Block>();
            for (int i = 0; i < 25; i++)
            {
                float width = rand.Next(10, 100);
                float height = rand.Next(10, 100);
                Blocks.Add(new Block(
                    new RectangleF(0, 0, width, height), height * 0.6f));
            }
        }

        private void picWriting_Resize(object sender, EventArgs e)
        {
            picWriting.Refresh();
        }

        // Flow and draw the blocks.
        private void picWriting_Paint(object sender, PaintEventArgs e)
        {
            e.Graphics.SmoothingMode = SmoothingMode.AntiAlias;

            // Flow.
            FlowBlocks(picWriting.ClientRectangle, Obstacles, Blocks);

            // Draw obstacles.
            foreach (RectangleF obstacle in Obstacles)
            {
                e.Graphics.FillRectangle(Brushes.Pink, obstacle);
                e.Graphics.DrawRectangle(Pens.Red, obstacle);
            }

            // Draw flowed blocks.
            for (int i = 0; i < Blocks.Count; i++)
            {
                Blocks[i].Draw(e.Graphics, i, picWriting.Font); 
            }
        }

        // Flow blocks around obstacles.
        private void FlowBlocks(RectangleF bounds,
            List<RectangleF> obstacles, List<Block> blocks)
        {
            float y = bounds.Y;

            // Repeat until we place all blocks or run out of room.
            int first_block = 0;
            while (first_block < blocks.Count)
            {
                // Position a row of blocks.
                int num_positioned = PositionOneRow(
                    bounds, obstacles, blocks,
                    ref first_block, ref y);

                // See if any fit.
                if (num_positioned == 0)
                {
                    // None fit. Move down.
                    MoveDown(bounds, obstacles, blocks[first_block], ref y);
                }

                // See if we have run out of room.
                if (y > bounds.Bottom)
                {
                    // Position the remaining blocks at (-1, -1).
                    for (int i = first_block; i < blocks.Count; i++)
                    {
                        blocks[i].Bounds.X = -1;
                        blocks[i].Bounds.Y = -1;
                    }
                    break;
                }
            }
        }

        // Return the maximum number of blocks that will fit
        // on one row starting at the indicated Y coordinate.
        // If we position any blocks, update first_block and y.
        private int PositionOneRow(RectangleF bounds,
            List<RectangleF> obstacles, List<Block> blocks,
            ref int first_block, ref float y)
        {
            // Loop through the blocks.
            int last_that_fits = blocks.Count - 1;
            for (int i = first_block; i < blocks.Count; i++)
            {
                // See if we can place blocks[first_block]
                // through blocks[i].
                if (!BlocksFit(bounds, obstacles, blocks, first_block, i, y))
                {
                    last_that_fits = i - 1;
                    break;
                }
            }

            // If no blocks fit, return 0.
            if (last_that_fits < first_block) return 0;

            // Position the blocks that fit again.
            BlocksFit(bounds, obstacles, blocks,
                first_block, last_that_fits, y);

            // Find the maximum y coordinate for these blocks.
            for (int i = first_block; i <= last_that_fits; i++)
            {
                if (y < blocks[i].Bounds.Bottom)
                    y = blocks[i].Bounds.Bottom;
            }

            // Update first_block.
            int num_blocks_positioned = last_that_fits - first_block + 1;
            first_block = last_that_fits + 1;

            // Return the number of blocks that fit.
            return num_blocks_positioned;
        }

        // Return true if the indicated blocks will
        // fit starting at the given Y coordinate.
        private bool BlocksFit(RectangleF bounds,
            List<RectangleF> obstacles, List<Block> blocks,
            int first_block, int last_block, float y)
        {
            // Find the maximum top height.
            float top_height = 0;
            for (int i = first_block; i <= last_block; i++)
            {
                if (top_height < blocks[i].TopHeight)
                    top_height = blocks[i].TopHeight;
            }

            // Set the baseline.
            float baseline = y + top_height;

            // Position the blocks.
            float x = bounds.X;
            for (int i = first_block; i <= last_block; i++)
            {
                Block block = blocks[i];
                block.Bounds.X = x;
                x += block.Bounds.Width;
                if (x > bounds.Right) return false;

                block.Bounds.Y = baseline - block.TopHeight;

                // See if the block intersects with an obstacle.
                RectangleF rect = block.Bounds;
                bool had_hit = true;
                while (had_hit)
                {
                    had_hit = false;
                    foreach (RectangleF obstacle in obstacles)
                    {
                        if (obstacle.IntersectsWith(rect))
                        {
                            had_hit = true;

                            // Move the block to the right.
                            rect.X = obstacle.Right;

                            // See if we are out of bounds.
                            if (rect.Right > bounds.Right) return false;
                        }
                    }
                }

                // Update the block's bounds.
                block.Bounds = rect;
                x = rect.Right;

                // Loop to position the next block.
            }

            // If we get this far, then we have
            // positioned all of the blocks.
            return true;
        }

        // Move down so the block will clear at least one new obstacle.
        private void MoveDown(RectangleF bounds,
            List<RectangleF> obstacles, Block block, ref float y)
        {
            float min_y = y + block.Bounds.Height;
            RectangleF rect = new RectangleF(
                bounds.X, y, bounds.Width, block.Bounds.Height);
            foreach (RectangleF obstacle in obstacles)
            {
                if (obstacle.IntersectsWith(rect))
                {
                    if (min_y > obstacle.Bottom)
                        min_y = obstacle.Bottom;
                }
            }

            y = min_y;
        }
    }
}
